@@ -189,5 +189,5 @@ class Cursor:
     ###
 
     def __checksock(self):
-        if self._connection is None:
+        if self._connection is None or not self._connection.open_socket:
             raise cql.ProgrammingError("Cursor has been closed.")
