@@ -15,7 +15,8 @@
 # limitations under the License.
 
 import cql
-from cql.marshal import int32_pack, int32_unpack, uint16_pack, uint16_unpack
+from cql.marshal import (int32_pack, int32_unpack, uint16_pack, uint16_unpack,
+                         int8_pack, int8_unpack)
 from cql.cqltypes import lookup_cqltype
 from cql.connection import Connection
 from cql.cursor import Cursor, _VOID_DESCRIPTION, _COUNT_DESCRIPTION
@@ -32,8 +33,6 @@ except ImportError:
 PROTOCOL_VERSION             = 0x01
 PROTOCOL_VERSION_MASK        = 0x7f
 
-# XXX: should these be called request/response instead? unclear which one will
-# apply if/when the server initiates streams in the other direction.
 HEADER_DIRECTION_FROM_CLIENT = 0x00
 HEADER_DIRECTION_TO_CLIENT   = 0x80
 HEADER_DIRECTION_MASK        = 0x80
@@ -85,14 +84,18 @@ class _MessageType(object):
                                  % (self.__class__.__name__, pname))
             setattr(self, pname, pval)
 
-    def send(self, f, streamid, compression=False):
+    def send(self, f, streamid, compression=None):
         body = StringIO()
         self.send_body(body)
         body = body.getvalue()
         version = PROTOCOL_VERSION | HEADER_DIRECTION_FROM_CLIENT
-        flags = 0 # no compression supported yet
+        flags = 0
+        if compression is not None and len(body) > 0:
+            body = compression(body)
+            flags |= 0x1
         msglen = int32_pack(len(body))
-        header = '%c%c%c%c%s' % (version, flags, streamid, self.opcode, msglen)
+        header = ''.join(map(int8_pack, (version, flags, streamid, self.opcode))) \
+                 + msglen
         f.write(header)
         if len(body) > 0:
             f.write(body)
@@ -102,18 +105,23 @@ class _MessageType(object):
         return '<%s(%s)>' % (self.__class__.__name__, ', '.join(paramstrs))
     __repr__ = __str__
 
-def read_frame(f):
+def read_frame(f, decompressor=None):
     header = f.read(8)
-    version, flags, stream, opcode = map(ord, header[:4])
+    version, flags, stream, opcode = map(int8_unpack, header[:4])
     body_len = int32_unpack(header[4:])
     assert version & PROTOCOL_VERSION_MASK == PROTOCOL_VERSION, \
             "Unsupported CQL protocol version %d" % version
     assert version & HEADER_DIRECTION_MASK == HEADER_DIRECTION_TO_CLIENT, \
             "Unexpected request from server with opcode %04x, stream id %r" % (opcode, stream)
     assert body_len >= 0, "Invalid CQL protocol body_len %r" % body_len
+    body = f.read(body_len)
+    if flags & 0x1:
+        if decompressor is None:
+            raise ProtocolException("No decompressor available for compressed frame!")
+        body = decompressor(body)
+        flags ^= 0x1
     if flags:
         warn("Unknown protocol flags set: %02x. May cause problems." % flags)
-    body = f.read(body_len)
     msgclass = _message_types_by_opcode[opcode]
     msg = msgclass.recv_body(StringIO(body))
     msg.stream_id = stream
@@ -380,7 +388,7 @@ class ResultMessage(_MessageType):
         return CqlResult(column_metadata=colspecs, rows=rows)
 
     @classmethod
-    def recv_results_prepared(self, f):
+    def recv_results_prepared(cls, f):
         queryid = read_int(f)
         colspecs = cls.recv_results_metadata(f)
         return (queryid, colspecs)
@@ -488,10 +496,10 @@ class EventMessage(_MessageType):
 
 
 def read_byte(f):
-    return ord(f.read(1))
+    return int8_unpack(f.read(1))
 
 def write_byte(f, b):
-    f.write(chr(b))
+    f.write(int8_pack(b))
 
 def read_int(f):
     return int32_unpack(f.read(4))
@@ -617,7 +625,8 @@ class NativeCursor(Cursor):
         return self._connection.wait_for_request(QueryMessage(query=query))
 
     def get_response_prepared(self, prepared_query, params):
-        em = ExecuteMessage(queryid=prepared_query.itemid, queryparams=params)
+        qparams = [params[pname] for pname in prepared_query.paramnames]
+        em = ExecuteMessage(queryid=prepared_query.itemid, queryparams=qparams)
         return self._connection.wait_for_request(em)
 
     def get_column_metadata(self, column_id):
@@ -670,10 +679,10 @@ class NativeCursor(Cursor):
         self.rowcount = len(self.result)
 
     def get_compression(self):
-        return None
+        return self._connection.compression
 
     def set_compression(self, val):
-        if val is not None:
+        if val != self.get_compression():
             raise NotImplementedError("Setting per-cursor compression is not "
                                       "supported in NativeCursor.")
 
@@ -702,6 +711,20 @@ class debugsock:
     def close(self):
         pass
 
+locally_supported_compressions = {}
+
+try:
+    import snappy
+except ImportError:
+    pass
+else:
+    # work around apparently buggy snappy decompress
+    def decompress(byts):
+        if byts == '\x00':
+            return ''
+        return snappy.decompress(byts)
+    locally_supported_compressions['snappy'] = (snappy.compress, decompress)
+
 class NativeConnection(Connection):
     cursorclass = NativeCursor
 
@@ -710,6 +733,8 @@ class NativeConnection(Connection):
         self.responses = {}
         self.waiting = {}
         self.conn_ready = False
+        self.compressor = self.decompressor = None
+        self.event_watchers = {}
         Connection.__init__(self, *args, **kwargs)
 
     def establish_connection(self):
@@ -721,7 +746,7 @@ class NativeConnection(Connection):
         self.open_socket = True
         supported = self.wait_for_request(OptionsMessage())
         self.supported_cql_versions = supported.cqlversions
-        self.supported_compressions = supported.options['COMPRESSION']
+        self.remote_supported_compressions = supported.options['COMPRESSION']
 
         if self.cql_version:
             if self.cql_version not in self.supported_cql_versions:
@@ -733,20 +758,30 @@ class NativeConnection(Connection):
             self.cql_version = self.supported_cql_versions[0]
 
         opts = {}
+        compresstype = None
         if self.compression:
-            if self.compression not in self.supported_compressions:
-                raise ProgrammingError("Compression type %r is not supported by"
-                                       " remote. Supported compression types: %r"
-                                       % (self.compression, self.supported_compressions))
-            # XXX: Remove this once some compressions are supported
-            raise NotImplementedError("CQL driver does not yet support compression")
-            opts['COMPRESSION'] = self.compression
+            overlap = set(locally_supported_compressions) \
+                    & set(self.remote_supported_compressions)
+            if len(overlap) == 0:
+                warn("No available compression types supported on both ends."
+                     " locally supported: %r. remotely supported: %r"
+                     % (locally_supported_compressions,
+                        self.remote_supported_compressions))
+            else:
+                compresstype = iter(overlap).next() # choose any
+                opts['COMPRESSION'] = compresstype
+                compr, decompr = locally_supported_compressions[compresstype]
+                # set the decompressor here, but set the compressor only after
+                # a successful Ready message
+                self.decompressor = decompr
 
         sm = StartupMessage(cqlversion=self.cql_version, options=opts)
         startup_response = self.wait_for_request(sm)
         while True:
             if isinstance(startup_response, ReadyMessage):
                 self.conn_ready = True
+                if compresstype:
+                    self.compressor = compr
                 break
             if isinstance(startup_response, AuthenticateMessage):
                 self.authenticator = startup_response.authenticator
@@ -761,8 +796,6 @@ class NativeConnection(Connection):
                 raise cql.InternalError("Unexpected response %r during connection setup"
                                         % startup_response)
 
-        if self.keyspace:
-            self.set_initial_keyspace(self.keyspace)
 
     def set_initial_keyspace(self, keyspace):
         c = self.cursor()
@@ -781,6 +814,11 @@ class NativeConnection(Connection):
 
         return self.wait_for_requests(msg)[0]
 
+    def send_msg(self, msg):
+        reqid = self.make_reqid()
+        msg.send(self.socketf, reqid, compression=self.compressor)
+        return reqid
+
     def wait_for_requests(self, *msgs):
         """
         Given any number of message objects, send them all to the server
@@ -791,9 +829,8 @@ class NativeConnection(Connection):
 
         reqids = []
         for msg in msgs:
-            reqid = self.make_reqid()
+            reqid = self.send_msg(msg)
             reqids.append(reqid)
-            msg.send(self.socketf, reqid)
         resultdict = self.wait_for_results(*reqids)
         return [resultdict[reqid] for reqid in reqids]
 
@@ -802,6 +839,10 @@ class NativeConnection(Connection):
         Given any number of stream-ids, wait until responses have arrived for
         each one, and return a dictionary mapping the stream-ids to the
         appropriate results.
+
+        For internal use, None may be passed in place of a reqid, which will
+        be considered satisfied when a message of any kind is received (and, if
+        appropriate, handled).
         """
 
         waiting_for = set(reqids)
@@ -815,12 +856,15 @@ class NativeConnection(Connection):
                 results[r] = result
                 waiting_for.remove(r)
         while waiting_for:
-            newmsg = read_frame(self.socketf)
+            newmsg = read_frame(self.socketf, decompressor=self.decompressor)
             if newmsg.stream_id in waiting_for:
                 results[newmsg.stream_id] = newmsg
                 waiting_for.remove(newmsg.stream_id)
             else:
                 self.handle_incoming(newmsg)
+            if None in waiting_for:
+                results[None] = newmsg
+                waiting_for.remove(None)
         return results
 
     def wait_for_result(self, reqid):
@@ -869,6 +913,81 @@ class NativeConnection(Connection):
         it may have to wait until something else waits on a result.
         """
 
-        reqid = self.make_reqid()
-        msg.send(self.socketf, reqid)
+        reqid = self.send_msg(msg)
         self.callback_when(reqid, cb)
+
+    def handle_pushed(self, msg):
+        """
+        Process an incoming message originated by the server.
+        """
+        watchers = self.event_watchers.get(msg.eventtype, ())
+        for cb in watchers:
+            cb(msg.eventargs)
+
+    def register_watcher(self, eventtype, cb):
+        """
+        Request that any events of the given type be passed to the given
+        callback when they arrive. Note that the callback may not be called
+        immediately upon the arrival of the event packet; it may have to wait
+        until something else waits on a result, or until wait_for_even() is
+        called.
+
+        If the event type has not been registered for already, this may
+        block while a new REGISTER message is sent to the server.
+
+        The available event types are in the cql.native.known_event_types
+        list.
+
+        When an event arrives, a dictionary will be passed to the callback
+        with the info about the event. Some example result dictionaries:
+
+        (For STATUS_CHANGE events:)
+
+          {'changetype': u'DOWN', 'address': ('12.114.19.76', 8000)}
+
+        (For TOPOLOGY_CHANGE events:)
+
+          {'changetype': u'NEW_NODE', 'address': ('19.10.122.13', 8000)}
+        """
+
+        if isinstance(eventtype, str):
+            eventtype = eventtype.decode('utf8')
+        try:
+            watchers = self.event_watchers[eventtype]
+        except KeyError:
+            ans = self.wait_for_request(RegisterMessage(eventlist=(eventtype,)))
+            if isinstance(ans, ErrorMessage):
+                raise cql.ProgrammingError("Server did not accept registration"
+                                           " for %s events: %s"
+                                           % (eventtype, ans.summarymsg()))
+            watchers = self.event_watchers.setdefault(eventtype, [])
+        watchers.append(cb)
+
+    def unregister_watcher(self, eventtype, cb):
+        """
+        Given an eventtype and a callback previously registered with
+        register_watcher(), remove that callback from the list of watchers for
+        the given event type.
+        """
+
+        if isinstance(eventtype, str):
+            eventtype = eventtype.decode('utf8')
+        self.event_watchers[eventtype].remove(cb)
+
+    def wait_for_event(self):
+        """
+        Wait for any sort of event to arrive, and handle it via the
+        registered callbacks. It is recommended that some event watchers
+        be registered before calling this; otherwise, no events will be
+        sent by the server.
+        """
+        eventsseen = []
+        def i_saw_an_event(ev):
+            eventsseen.append(ev)
+        wlists = self.event_watchers.values()
+        for wlist in wlists:
+            wlist.append(i_saw_an_event)
+        while not eventsseen:
+            self.wait_for_result(None)
+        for wlist in wlists:
+            wlist.remove(i_saw_an_event)
